from lueur.make_id import make_id
from lueur.models import AWSMeta, Resource
from lueur.platform.aws.client import Client

__all__ = ["explore_ec2"]


def explore_ec2(region: str) -> list[Resource]:
    resources = []

    instances = explore_instances(region)
    resources.extend(instances)

    return resources


###############################################################################
# Private functions
###############################################################################
def explore_instances(region: str) -> list[Resource]:
    results = []

    with Client("ec2", region) as c:
        instances = c.describe_instances()

        for reservations in instances["Reservations"]:
            for instance in reservations["Instances"]:
                results.append(
                    Resource(
                        id=make_id(instance["InstanceId"]),
                        meta=AWSMeta(
                            name=instance["InstanceId"],
                            display=instance["InstanceId"],
                            kind="instance",
                            platform="aws",
                            region=region,
                            category="compute",
                        ),
                        struct=instance,
                    )
                )

    return results



from typing import Any, cast

import jsonpath

from lueur.models import Discovery, Link

__all__ = ["add_link", "get_linked_resource"]


def add_link(d: Discovery, resource_id: str, link: Link) -> None:
    for r in d.resources:
        if r.id == resource_id:
            r.links.append(link)
            break


def get_linked_resource(
    discovery: dict[str, Any], pointer: str
) -> dict[str, str]:
    return cast(
        dict[str, str], jsonpath.JSONPointer(pointer).resolve(discovery)
    )



from concurrent.futures import ThreadPoolExecutor

from lueur.make_id import make_id
from lueur.models import AWSMeta, Resource
from lueur.platform.aws.client import Client

__all__ = ["explore_ecr"]


def explore_ecr(region: str) -> list[Resource]:
    resources = []

    repositories = explore_repositories(region)
    resources.extend(repositories)

    if not repositories:
        return resources

    futures = []
    with ThreadPoolExecutor(max_workers=len(repositories)) as executor:
        for c in repositories:
            registry_id = c.struct["registryId"]
            repository_name = c.struct["repositoryName"]

            futures.append(
                executor.submit(
                    explore_images, region, registry_id, repository_name
                )
            )

    for future in futures:
        resources.extend(future.result())

    return resources


###############################################################################
# Private functions
###############################################################################
def explore_repositories(region: str) -> list[Resource]:
    results = []

    with Client("ecr", region) as c:
        repositories = c.describe_repositories()

        for repository in repositories["repositories"]:
            results.append(
                Resource(
                    id=make_id(repository["repositoryArn"]),
                    meta=AWSMeta(
                        name=repository["repositoryName"],
                        display=repository["repositoryName"],
                        kind="repository",
                        platform="aws",
                        region=region,
                        category="compute",
                    ),
                    struct=repository,
                )
            )

    return results


def explore_images(
    region: str, registry_id: str, repository_name: str
) -> list[Resource]:
    with Client("ecr", region) as c:
        images = c.describe_images(
            registryId=registry_id, repositoryName=repository_name
        )

    results = []
    for image in images["imageDetails"]:
        name = f"{image['registryId']}/{image['repositoryName']}"
        results.append(
            Resource(
                id=make_id(image["imageDigest"]),
                meta=AWSMeta(
                    name=name,
                    display=name,
                    kind="image",
                    platform="aws",
                    region=region,
                    category="storage",
                ),
                struct=image,
            )
        )

    return results



from typing import Any

import msgspec
import pytest
import respx

from lueur.platform.gcp.securities import explore_securities

from tests.platform.gcp.utils import patch_auth


@respx.mock
@pytest.mark.asyncio
async def test_explore_global_securities(
    gcp_project: str, gcp_global_securities: list[dict[str, Any]]
) -> None:
    respx.post(
        "https://oauth2.googleapis.com/token",
    ).respond(content=msgspec.json.encode({"access_token": "xyz"}))

    respx.get(
        f"https://compute.googleapis.com/compute/v1/projects/{gcp_project}/global/securityPolicies",
    ).respond(content=msgspec.json.encode({"items": gcp_global_securities}))

    with patch_auth():
        resources = await explore_securities(gcp_project)
    assert len(resources) == 1


@respx.mock
@pytest.mark.asyncio
async def test_explore_regional_securities(
    gcp_project: str,
    gcp_region: str,
) -> None:
    respx.post(
        "https://oauth2.googleapis.com/token",
    ).respond(content=msgspec.json.encode({"access_token": "xyz"}))

    respx.get(
        f"https://compute.googleapis.com/compute/v1/projects/{gcp_project}/regions/{gcp_region}/securityPolicies",
    ).respond(content=msgspec.json.encode({"items": []}))

    with patch_auth():
        resources = await explore_securities(gcp_project, gcp_region)

    assert len(resources) == 0



# mypy: disable-error-code="index,union-attr"
import asyncio
import logging

import msgspec
from google.oauth2._service_account_async import Credentials

from lueur.make_id import make_id
from lueur.models import GCPMeta, Resource
from lueur.platform.gcp.client import AuthorizedSession, Client

__all__ = ["explore_dns"]
logger = logging.getLogger("lueur.lib")


async def explore_dns(
    project: str, creds: Credentials | None = None
) -> list[Resource]:
    resources = []

    async with Client("https://dns.googleapis.com", creds) as c:
        zones = await explore_managedzones(c, project)

        tasks: list[asyncio.Task] = []

        async with asyncio.TaskGroup() as tg:
            for zone in zones:
                tasks.append(
                    tg.create_task(
                        explore_recordsets(c, project, zone.meta.name)
                    )
                )

        for task in tasks:
            result = task.result()
            if result is None:
                continue
            resources.extend(result)

    return resources


###############################################################################
# Private functions
###############################################################################
async def explore_managedzones(
    c: AuthorizedSession, project: str
) -> list[Resource]:
    response = await c.get(f"/dns/v1/projects/{project}/managedZones")

    managedZones = msgspec.json.decode(response.content)

    if response.status_code == 403:
        logger.warning(f"DNS API access failure: {managedZones}")
        return []

    if "rrsets" not in managedZones:
        logger.warning(f"No managedZones found: {managedZones}")
        return []

    results = []
    for managedZone in managedZones["rrsets"]:
        results.append(
            Resource(
                id=make_id(managedZone["id"]),
                meta=GCPMeta(
                    name=managedZone["name"],
                    display=managedZone["dnsName"],
                    kind="managedzone",
                    platform="gcp",
                    project=project,
                    category="dns",
                ),
                struct=managedZone,
            )
        )

    return results


async def explore_recordsets(
    c: AuthorizedSession, project: str, zone: str
) -> list[Resource]:
    response = await c.get(
        f"/dns/v1/projects/{project}/managedZones/{zone}/rrsets"
    )

    recordsets = msgspec.json.decode(response.content)

    if response.status_code == 403:
        logger.warning(f"DNS API access failure: {recordsets}")
        return []

    if "rrsets" not in recordsets:
        logger.warning(f"No recordsets found: {recordsets}")
        return []

    results = []
    for recordset in recordsets["rrsets"]:
        results.append(
            Resource(
                id=make_id(recordset["name"]),
                meta=GCPMeta(
                    name=recordset["name"],
                    display=recordset["name"],
                    kind="recordset",
                    platform="gcp",
                    project=project,
                    category="dns",
                ),
                struct=recordset,
            )
        )

    return results



# mypy: disable-error-code="union-attr"
import asyncio
import logging

import msgspec
from google.oauth2._service_account_async import Credentials

from lueur.make_id import make_id
from lueur.models import GCPMeta, Resource
from lueur.platform.gcp.client import AuthorizedSession, Client

__all__ = ["explore_target_proxies"]
logger = logging.getLogger("lueur.lib")


async def explore_target_proxies(
    project: str, location: str | None = None, creds: Credentials | None = None
) -> list[Resource]:
    resources = []

    tasks: list[asyncio.Task] = []

    async with Client("https://compute.googleapis.com", creds) as c:
        async with asyncio.TaskGroup() as tg:
            if location:
                tasks.append(
                    tg.create_task(
                        explore_regional_http_target_proxies(
                            c, project, location
                        )
                    )
                )
                tasks.append(
                    tg.create_task(
                        explore_regional_https_target_proxies(
                            c, project, location
                        )
                    )
                )
            else:
                tasks.append(
                    tg.create_task(
                        explore_global_http_target_proxies(c, project)
                    )
                )
                tasks.append(
                    tg.create_task(
                        explore_global_https_target_proxies(c, project)
                    )
                )

    for task in tasks:
        result = task.result()
        if result is None:
            continue
        resources.extend(result)

    return resources


###############################################################################
# Private functions
###############################################################################
async def explore_global_http_target_proxies(
    c: AuthorizedSession, project: str
) -> list[Resource]:
    response = await c.get(
        f"/compute/v1/projects/{project}/global/targetHttpProxies"
    )

    proxies = msgspec.json.decode(response.content)

    if response.status_code == 403:
        logger.warning(f"Target Proxy API access failure: {proxies}")
        return []

    if "items" not in proxies:
        logger.warning(f"No global HTTP proxies found: {proxies}")
        return []

    results = []
    for proxy in proxies.get("items", []):
        self_link = proxy["selfLink"]
        name = proxy["name"]
        display = name

        results.append(
            Resource(
                id=make_id(proxy["id"]),
                meta=GCPMeta(
                    name=name,
                    display=display,
                    kind="global-target-http-proxies",
                    project=project,
                    self_link=self_link,
                    platform="gcp",
                    category="loadbalancer",
                ),
                struct=proxy,
            )
        )

    return results


async def explore_regional_http_target_proxies(
    c: AuthorizedSession, project: str, location: str
) -> list[Resource]:
    response = await c.get(
        f"/compute/v1/projects/{project}/regions/{location}/targetHttpProxies"
    )

    proxies = msgspec.json.decode(response.content)

    if response.status_code == 403:
        logger.warning(f"Target Proxy API access failure: {proxies}")
        return []

    if "items" not in proxies:
        logger.warning(f"No regional HTTP proxies found: {proxies}")
        return []

    results = []
    for proxy in proxies.get("items", []):
        self_link = proxy["selfLink"]
        name = proxy["name"]
        display = name

        results.append(
            Resource(
                id=make_id(proxy["id"]),
                meta=GCPMeta(
                    name=name,
                    display=display,
                    kind="regional-target-http-proxies",
                    project=project,
                    region=location,
                    self_link=self_link,
                    platform="gcp",
                    category="loadbalancer",
                ),
                struct=proxy,
            )
        )

    return results


async def explore_global_https_target_proxies(
    c: AuthorizedSession, project: str
) -> list[Resource]:
    response = await c.get(
        f"/compute/v1/projects/{project}/global/targetHttpsProxies"
    )

    proxies = msgspec.json.decode(response.content)

    if response.status_code == 403:
        logger.warning(f"Target Proxy API access failure: {proxies}")
        return []

    if "items" not in proxies:
        logger.warning(f"No global HTTPs proxies found: {proxies}")
        return []

    results = []
    for proxy in proxies.get("items", []):
        self_link = proxy["selfLink"]
        name = proxy["name"]
        display = name

        results.append(
            Resource(
                id=make_id(proxy["id"]),
                meta=GCPMeta(
                    name=name,
                    display=display,
                    kind="global-target-https-proxies",
                    project=project,
                    self_link=self_link,
                    platform="gcp",
                    category="loadbalancer",
                ),
                struct=proxy,
            )
        )

    return results


async def explore_regional_https_target_proxies(
    c: AuthorizedSession, project: str, location: str
) -> list[Resource]:
    response = await c.get(
        f"/compute/v1/projects/{project}/regions/{location}/targetHttpsProxies"
    )

    proxies = msgspec.json.decode(response.content)

    if response.status_code == 403:
        logger.warning(f"Target Proxy API access failure: {proxies}")
        return []

    if "items" not in proxies:
        logger.warning(f"No regional HTTPs proxies found: {proxies}")
        return []

    results = []
    for proxy in proxies.get("items", []):
        self_link = proxy["selfLink"]
        name = proxy["name"]
        display = name

        results.append(
            Resource(
                id=make_id(proxy["id"]),
                meta=GCPMeta(
                    name=name,
                    display=display,
                    kind="regional-target-https-proxies",
                    project=project,
                    region=location,
                    self_link=self_link,
                    platform="gcp",
                    category="loadbalancer",
                ),
                struct=proxy,
            )
        )

    return results



from typing import Any

import msgspec
import pytest
import respx

from lueur.platform.gcp.gke import explore_gke

from tests.platform.gcp.utils import patch_auth


@respx.mock
@pytest.mark.asyncio
async def test_explore_connector(
    gcp_project: str,
    gcp_region: str,
    gcp_gke_cluster_name: str,
    gcp_gke_clusters: list[dict[str, Any]],
    gcp_gke_cluster_nodepools: list[dict[str, Any]],
) -> None:
    respx.post(
        "https://oauth2.googleapis.com/token",
    ).respond(content=msgspec.json.encode({"access_token": "xyz"}))

    respx.get(
        f"https://container.googleapis.com/v1/projects/{gcp_project}/locations/{gcp_region}/clusters",
    ).respond(content=msgspec.json.encode({"clusters": gcp_gke_clusters}))

    respx.get(
        f"https://container.googleapis.com/v1/projects/{gcp_project}/locations/{gcp_region}/clusters/{gcp_gke_cluster_name}/nodePools",
    ).respond(
        content=msgspec.json.encode({"nodePools": gcp_gke_cluster_nodepools})
    )

    with patch_auth():
        resources = await explore_gke(gcp_project, gcp_region)



import asyncio
import logging

import msgspec
from google.oauth2._service_account_async import Credentials

from lueur.make_id import make_id
from lueur.models import GCPMeta, Resource
from lueur.platform.gcp.client import AuthorizedSession, Client

__all__ = ["explore_connector"]
logger = logging.getLogger("lueur.lib")


async def explore_connector(
    project: str, location: str, creds: Credentials | None = None
) -> list[Resource]:
    resources = []

    async with Client("https://connectors.googleapis.com", creds) as c:
        providers = await explore_connector_providers(c, project, location)
        resources.extend(providers)

        connections = await explore_connections(c, project, location)
        resources.extend(connections)

        tasks: list[asyncio.Task] = []

        async with asyncio.TaskGroup() as tg:
            for p in providers:
                tasks.append(
                    tg.create_task(
                        explore_connectors(c, project, location, p.meta.name)
                    )
                )

        for task in tasks:
            result = task.result()
            if result is None:
                continue
            resources.extend(result)

    return resources


###############################################################################
# Private functions
###############################################################################
async def explore_connector_providers(
    c: AuthorizedSession, project: str, location: str
) -> list[Resource]:
    response = await c.get(
        f"/v1/projects/{project}/locations/{location}/providers"
    )

    providers = msgspec.json.decode(response.content)

    if response.status_code == 403:
        logger.warning(f"Connectors providers API access failure: {providers}")
        return []

    if "providers" not in providers:
        logger.warning(f"No connectors providers found: {providers}")
        return []

    results = []
    for provider in providers.get("providers", []):
        name = provider["name"]
        display = provider["displayName"]
        self_link = provider.get("selfLink")

        results.append(
            Resource(
                id=make_id(provider["name"]),
                meta=GCPMeta(
                    name=name,
                    display=display,
                    kind="connector-provider",
                    project=project,
                    region=location,
                    platform="k8s",
                    self_link=self_link,
                    category="integration",
                ),
                struct=provider,
            )
        )

    return results


async def explore_connectors(
    c: AuthorizedSession, project: str, location: str, name: str
) -> list[Resource]:
    response = await c.get(
        f"/v1/projects/{project}/locations/{location}/providers/{name}/connectors"
    )

    connectors = msgspec.json.decode(response.content)

    if response.status_code == 403:
        logger.warning(f"Connectors API access failure: {connectors}")
        return []

    if "connectors" not in connectors:
        logger.warning(f"No connectors found: {connectors}")
        return []

    results = []
    for connector in connectors.get("connectors", []):
        name = connector["name"]
        display = connector["displayName"]
        self_link = connector.get("selfLink")

        results.append(
            Resource(
                id=make_id(connector["name"]),
                meta=GCPMeta(
                    name=name,
                    display=display,
                    kind="connector",
                    project=project,
                    region=location,
                    platform="gcp",
                    self_link=self_link,
                    category="integration",
                ),
                struct=connector,
            )
        )

    return results


async def explore_connections(
    c: AuthorizedSession, project: str, location: str
) -> list[Resource]:
    response = await c.get(
        f"/v1/projects/{project}/locations/{location}/connections"
    )

    connections = msgspec.json.decode(response.content)

    if response.status_code == 403:
        logger.warning(f"Connectors API access failure: {connections}")
        return []

    if "connections" not in connections:
        logger.warning(f"No connections found: {connections}")
        return []

    results = []
    for connection in connections.get("connections", []):
        name = connection["name"]
        display = connection["name"]
        self_link = connection.get("selfLink")

        results.append(
            Resource(
                id=make_id(connection["name"]),
                meta=GCPMeta(
                    name=name,
                    display=display,
                    kind="connector-connection",
                    project=project,
                    region=location,
                    platform="gcp",
                    self_link=self_link,
                    category="integration",
                ),
                struct=connection,
            )
        )

    return results



from typing import Any

import msgspec
import pytest
import respx

from lueur.platform.gcp.vpc import explore_vpc

from tests.platform.gcp.utils import patch_auth


@respx.mock
@pytest.mark.asyncio
async def test_explore_vpc(
    gcp_project: str,
    gcp_region: str,
    gcp_vpc_networks: list[dict[str, Any]],
    gcp_vpc_subnets: list[dict[str, Any]],
) -> None:
    respx.post(
        "https://oauth2.googleapis.com/token",
    ).respond(content=msgspec.json.encode({"access_token": "xyz"}))

    respx.get(
        f"https://compute.googleapis.com/compute/v1/projects/{gcp_project}/global/networks",
    ).respond(content=msgspec.json.encode({"items": gcp_vpc_networks}))

    respx.get(
        f"https://www.googleapis.com/compute/v1/projects/{gcp_project}/regions/{gcp_region}/subnetworks/{gcp_project}-subnet",
    ).respond(content=msgspec.json.encode(gcp_vpc_subnets[2]))

    respx.get(
        f"https://www.googleapis.com/compute/v1/projects/{gcp_project}/regions/{gcp_region}/subnetworks/{gcp_project}-demo-subnet",
    ).respond(content=msgspec.json.encode(gcp_vpc_subnets[1]))

    with patch_auth():
        resources = await explore_vpc(gcp_project)
    assert len(resources) == 4



# mypy: disable-error-code="union-attr"
import logging
from typing import Any

import msgspec
from google.oauth2._service_account_async import Credentials

from lueur.links import add_link
from lueur.make_id import make_id
from lueur.models import Discovery, GCPMeta, Link, Resource
from lueur.platform.gcp.client import AuthorizedSession, Client
from lueur.rules import iter_resource

__all__ = ["explore_cloudrun"]
logger = logging.getLogger("lueur.lib")


async def explore_cloudrun(
    project: str, location: str, creds: Credentials | None = None
) -> list[Resource]:
    resources = []

    async with Client("https://run.googleapis.com", creds) as c:
        services = await explore_services(c, project, location)
        resources.extend(services)

    async with Client("https://vpcaccess.googleapis.com", creds) as c:
        connectors = await explore_vpcaccess_connectors(c, project, location)
        resources.extend(connectors)

    return resources


###############################################################################
# Private functions
###############################################################################
async def explore_services(
    c: AuthorizedSession, project: str, location: str
) -> list[Resource]:
    response = await c.get(
        f"/v2/projects/{project}/locations/{location}/services"
    )

    services = msgspec.json.decode(response.content)

    if response.status_code == 403:
        logger.warning(f"Services API access failure: {services}")
        return []

    if "services" not in services:
        logger.warning(f"No services found: {services}")
        return []

    results = []
    for svc in services.get("services", []):
        name = svc["name"]
        display = name.rsplit("/", 1)[-1]
        self_link = svc.get("selfLink")

        results.append(
            Resource(
                id=make_id(svc["uid"]),
                meta=GCPMeta(
                    name=name,
                    display=display,
                    kind="cloudrun",
                    project=project,
                    region=location,
                    platform="gcp",
                    self_link=self_link,
                    category="compute",
                ),
                struct=svc,
            )
        )

    return results


async def explore_vpcaccess_connectors(
    c: AuthorizedSession, project: str, location: str
) -> list[Resource]:
    response = await c.get(
        f"/v1beta1/projects/{project}/locations/{location}/connectors"
    )

    connectors = msgspec.json.decode(response.content)

    if response.status_code == 403:
        logger.warning(f"VPC connectors API access failure: {connectors}")
        return []

    if "connectors" not in connectors:
        logger.warning(f"No VPC access connectors found: {connectors}")
        return []

    results = []
    for connector in connectors["connectors"]:
        name = connector["name"]
        display = name.rsplit("/", 1)[-1]
        self_link = connector.get("selfLink")

        results.append(
            Resource(
                id=make_id(connector["name"]),
                meta=GCPMeta(
                    name=name,
                    display=display,
                    kind="cloudrun-vpcaccess-connector",
                    project=project,
                    region=location,
                    platform="gcp",
                    self_link=self_link,
                    category="network",
                ),
                struct=connector,
            )
        )

    return results


def expand_links(d: Discovery, serialized: dict[str, Any]) -> None:
    for s in iter_resource(
        serialized, "$.resources[?@.meta.kind=='cloudrun'].meta.name"
    ):
        r_id = s.parent.parent.obj["id"]  # type: ignore
        name = s.value.rsplit("/", 1)[-1]  # type: ignore
        p = f"$.resources[?@.meta.kind=='service' && @.struct.cloudRun.serviceName=='{name}']"  # noqa E501
        for svc in iter_resource(serialized, p):
            add_link(
                d,
                r_id,
                Link(
                    direction="out",
                    kind="service",
                    path=svc.path,
                    pointer=str(svc.pointer()),
                ),
            )

            svc_name = svc.obj["meta"]["name"]  # type: ignore

            p = f"$.resources[?@.meta.kind=='slo' && match(@.meta.name, '{svc_name}/serviceLevelObjectives/.*')]"  # noqa E501
            for slo in iter_resource(serialized, p):
                add_link(
                    d,
                    r_id,
                    Link(
                        direction="out",
                        kind="slo",
                        path=slo.path,
                        pointer=str(slo.pointer()),
                    ),
                )

        p = f"$.resources[?@.meta.kind=='regional-neg' && @.struct.cloudRun.service=='{name}']"  # noqa E501
        for neg in iter_resource(serialized, p):
            add_link(
                d,
                r_id,
                Link(
                    direction="out",
                    kind="regional-neg",
                    path=neg.path,
                    pointer=str(neg.pointer()),
                ),
            )

        tpl = s.parent.parent.obj["struct"]["template"]  # type: ignore

        connector_name = tpl.get("vpcAccess", {}).get("connector")
        if connector_name:
            p = f"$.resources[?@.meta.kind=='cloudrun-vpcaccess-connector' && @.meta.name=='{connector_name}']"  # noqa E501
            for svc in iter_resource(serialized, p):
                add_link(
                    d,
                    r_id,
                    Link(
                        direction="out",
                        kind="cloudrun-vpcaccess-connector",
                        path=svc.path,
                        pointer=str(svc.pointer()),
                    ),
                )

        nics = tpl.get("vpcAccess", {}).get("networkInterfaces")
        if nics:
            for nic in nics:
                network = nic["network"]
                p = f"$.resources[?@.meta.kind=='network' && @.meta.name=='{network}']"  # noqa E501
                for nk in iter_resource(serialized, p):
                    add_link(
                        d,
                        r_id,
                        Link(
                            direction="out",
                            kind="network",
                            path=nk.path,
                            pointer=str(nk.pointer()),
                        ),
                    )

                subnet = nic["subnetwork"]
                p = f"$.resources[?@.meta.kind=='subnet' && @.meta.name=='{subnet}']"  # noqa E501
                for sn in iter_resource(serialized, p):
                    add_link(
                        d,
                        r_id,
                        Link(
                            direction="out",
                            kind="subnet",
                            path=sn.path,
                            pointer=str(sn.pointer()),
                        ),
                    )

        for v in tpl.get("volumes", []):
            if v.get("name") != "cloudsql":
                continue

            for inst in v["cloudSqlInstance"].get("instances", []):
                p = f"$.resources[?@.meta.kind=='instance' && @.struct.connectionName=='{inst}']"  # noqa E501
                for svc in iter_resource(serialized, p):
                    add_link(
                        d,
                        r_id,
                        Link(
                            direction="out",
                            kind="instance",
                            path=svc.path,
                            pointer=str(svc.pointer()),
                        ),
                    )

    for s in iter_resource(
        serialized,
        "$.resources[?@.meta.kind=='cloudrun-vpcaccess-connector'].struct.network",  # noqa E501
    ):
        r_id = s.parent.parent.obj["id"]  # type: ignore
        network = s.value.rsplit("/", 1)[-1]  # type: ignore
        p = f"$.resources[?@.meta.kind=='network' && @.meta.name=='{network}']"
        for svc in iter_resource(serialized, p):
            add_link(
                d,
                r_id,
                Link(
                    direction="out",
                    kind="network",
                    path=svc.path,
                    pointer=str(svc.pointer()),
                ),
            )



from contextlib import contextmanager
from typing import Any, Callable, Generator
from unittest.mock import patch, AsyncMock

from google.auth.credentials import AnonymousCredentials


class TestAnonymousCredentials(AnonymousCredentials):
    async def before_request(self, request, method, url, headers):
        pass


@contextmanager
def patch_auth() -> Generator[AsyncMock, None, None]:
    with patch("lueur.platform.gcp.client.default_async", autospec=True) as p:
        p.return_value = (TestAnonymousCredentials(), "")
        yield p



from textwrap import dedent, wrap

__all__ = ["collapse"]


def collapse(text: str) -> str:
    """
    Collpase a multiline string into a single line string.
    """
    return " ".join(wrap(dedent(text))).strip()



from collections.abc import AsyncIterator
from contextlib import asynccontextmanager
from typing import Any

import httpx
from google.auth import exceptions, transport
from google.auth._credentials_async import Credentials
from google.auth._default_async import default_async
from google.oauth2._service_account_async import Credentials as OAuthCreds
from tenacity import (
    retry,
    retry_if_exception_type,
    stop_after_attempt,
    wait_exponential,
)

__all__ = ["Client", "AuthorizedSession"]


class _Response(transport.Response):
    """
    Requests transport response adapter.
    Args:
        response (httpx.Response): The raw Requests response.
    """

    def __init__(self, response: httpx.Response):
        self._response = response
        self._content: bytes | None = None

    @property
    def status(self) -> int:
        return self._response.status_code

    @property
    def headers(self) -> httpx.Headers:
        return self._response.headers

    @property
    def data(self) -> bytes:
        return self._response.content

    async def raw_content(self):
        return await self.content()

    async def content(self) -> bytes:
        if self._content is None:
            self._content = await self._response.aread()
        return self._content


class Request(transport.Request):
    def __init__(self, httpx_client: httpx.AsyncClient) -> None:
        self.client = httpx_client

    async def __call__(
        self,
        url: str,
        method: str = "GET",
        body: Any = None,
        headers: dict[str, str] | None = None,
        timeout: httpx._types.TimeoutTypes | None = None,
        **kwargs,
    ):
        try:
            response = await self.client.request(
                method,
                url,
                data=body,
                headers=headers,
                timeout=timeout or httpx.Timeout(60),
                **kwargs,
            )
            return _Response(response)

        except httpx.HTTPError as caught_exc:
            new_exc = exceptions.TransportError(caught_exc)
            raise new_exc from caught_exc


class AuthorizedSession(httpx.AsyncClient):
    def __init__(
        self,
        credentials: Credentials | OAuthCreds,
        **kwargs,
    ):
        super(AuthorizedSession, self).__init__(**kwargs)
        self.credentials: Credentials = credentials

    @retry(
        retry=(
            retry_if_exception_type(httpx.TimeoutException)
            | retry_if_exception_type(httpx.ConnectError)
        ),
        stop=stop_after_attempt(3),
        wait=wait_exponential(multiplier=1, min=5, max=20),
    )
    async def request(  # type: ignore
        self,
        method: str,
        url: str,
        data: dict[str, str] | None = None,
        headers: dict[str, str] | None = None,
        **kwargs,
    ) -> httpx.Response:
        limits = httpx.Limits(max_connections=10, max_keepalive_connections=10)
        timeout = httpx.Timeout(60)

        async with httpx.AsyncClient(http2=True, limits=limits) as session:
            request_headers = headers.copy() if headers is not None else {}

            await self.credentials.before_request(
                Request(session), method, url, request_headers
            )

            # we want to force or timeout value
            kwargs.pop("timeout", None)

            response: httpx.Response = await super(
                AuthorizedSession, self
            ).request(
                method,
                url,
                data=data,
                headers=request_headers,
                timeout=timeout,
                **kwargs,
            )

            return response


@asynccontextmanager
async def Client(
    base_url: str, creds: OAuthCreds | None = None
) -> AsyncIterator[AuthorizedSession]:
    credentials = creds
    if creds is None:
        credentials, _ = default_async()

    async with AuthorizedSession(
        credentials,  # type: ignore
        base_url=httpx.URL(base_url),
    ) as s:
        yield s



import asyncio
from collections.abc import AsyncIterator
from contextlib import asynccontextmanager
from typing import Any, Type

from kubernetes import client, config
from urllib3.response import HTTPResponse

__all__ = ["Client", "AsyncClient"]


class AsyncClient:
    def __init__(self, api: Type) -> None:
        self.api: Type | None = api
        self.client: client.ApiClient | None = None
        self.inst = None

    async def __aenter__(self) -> "AsyncClient":
        assert self.api

        config.load_kube_config()
        self.client = client.ApiClient()
        self.inst = self.api(self.client)

        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None = None,
        exc_value: BaseException | None = None,
        traceback: Any | None = None,
    ) -> None:
        self.api = None
        self.client = None
        self.inst = None

    async def execute(
        self, func: str, /, *args, **kwargs
    ) -> HTTPResponse | client.ApiException:
        def _run(*args, **kwargs) -> HTTPResponse:
            f = getattr(self.inst, func)
            try:
                return f(*args, _preload_content=False, **kwargs)
            except client.ApiException as x:
                return x

        return await asyncio.to_thread(_run, *args, **kwargs)


@asynccontextmanager
async def Client(api: Type) -> AsyncIterator[AsyncClient]:
    async with AsyncClient(api) as c:
        yield c



# mypy: disable-error-code="call-arg"
import logging
from typing import Any

import msgspec
from kubernetes import client

from lueur.links import add_link
from lueur.make_id import make_id
from lueur.models import Discovery, K8SMeta, Link, Resource
from lueur.platform.k8s.client import AsyncClient, Client
from lueur.rules import iter_resource

__all__ = ["explore_replicaset"]
logger = logging.getLogger("lueur.lib")


async def explore_replicaset() -> list[Resource]:
    resources = []

    async with Client(client.AppsV1Api) as c:
        rs = await explore_replicasets(c)
        resources.extend(rs)

    return resources


###############################################################################
# Private functions
###############################################################################
async def explore_replicasets(c: AsyncClient) -> list[Resource]:
    response = await c.execute("list_replica_set_for_all_namespaces")

    replicasets = msgspec.json.decode(response.data)

    if response.status_code == 403:  # type: ignore
        logger.warning(f"K8S API server access failure: {replicasets}")
        return []

    if "items" not in replicasets:
        logger.warning(f"No replicasets found: {replicasets}")
        return []

    results = []
    for rs in replicasets["items"]:
        meta = rs["metadata"]
        results.append(
            Resource(
                id=make_id(meta["uid"]),
                meta=K8SMeta(
                    name=meta["name"],
                    display=meta["name"],
                    kind="replicaset",
                    platform="k8s",
                    ns=meta["namespace"],
                    category="compute",
                ),
                struct=rs,
            )
        )

    return results


def expand_links(d: Discovery, serialized: dict[str, Any]) -> None:
    for rs in iter_resource(
        serialized,
        "$.resources[?@.meta.kind=='replicaset' && @.meta.platform=='k8s'].meta.name",  # noqa: E501
    ):
        r_id = rs.parent.parent.obj["id"]  # type: ignore
        name = rs.value

        p = f"$.resources[?@.meta.kind=='pod'].struct.metadata.ownerReferences[?@.kind=='ReplicaSet' && @.name=='{name}']"  # noqa E501
        for ownerRef in iter_resource(serialized, p):
            pod = ownerRef.parent.parent.parent.parent  # type: ignore
            add_link(
                d,
                r_id,
                Link(
                    direction="out",
                    kind="pod",
                    path=pod.path,  # type: ignore
                    pointer=str(pod.pointer()),  # type: ignore
                ),
            )



import msgspec
from google.oauth2._service_account_async import Credentials

from lueur.make_id import make_id
from lueur.models import GCPProject
from lueur.platform.gcp.client import Client

__all__ = ["list_all_projects", "list_organization_projects"]


async def list_organization_projects(
    organization_id: str, creds: Credentials | None = None
) -> list[str]:
    async with Client(
        "https://cloudresourcemanager.googleapis.com", creds
    ) as c:
        response = await c.get(
            "/v3/projects",
            params={"parent": f"organizations/{organization_id}"},
        )

        projects = msgspec.json.decode(response.content)

        return [p["name"] for p in projects["projects"]]


async def list_all_projects(
    creds: Credentials | None = None,
) -> list[GCPProject]:
    async with Client(
        "https://cloudresourcemanager.googleapis.com", creds
    ) as c:
        response = await c.get(
            "/v3/projects:search", params={"query": "state:ACTIVE"}
        )

        projects = msgspec.json.decode(response.content)

        result = []

        result.extend(
            [
                GCPProject(
                    organization=p["parent"],
                    id=make_id(p["name"]),
                    name=p["displayName"],
                )
                for p in projects["projects"]
            ]
        )

        return result
